﻿/*
 * Copyright (c) 2010-2018 by Daniel Ranft, Wolfgang Hardt, Alexander Strobel
 *   Giegerich & Partner GmbH, Dreieich, Germany
 * Licensed under the MIT License, see LICENSE file for details.
 *
 *
 * Sourcecode: https://www.giepa.de/produkte/gpg4o/gnupgwrapper/
 */
using System;
using System.Runtime.InteropServices;
using System.Threading;

namespace GupGPGWrapper
{
	internal class WaitHelper
	{
		public const Int32 INFINITE = -1;
		const UInt32 WAIT_ABANDONED = 0x00000080;
		const UInt32 WAIT_ABANDONED_0 = WAIT_ABANDONED;
		const UInt32 WAIT_OBJECT_0 = 0x00000000;
		const UInt32 WAIT_TIMEOUT = 0x00000102;

		[DllImport("kernel32.dll", SetLastError = true)]
		static extern UInt32 WaitForSingleObject(IntPtr hHandle, UInt32 dwMilliseconds);

		[DllImport("kernel32.dll", SetLastError = true)]
		static extern int WaitForMultipleObjects(int nCount, IntPtr[] lpHandles, int bWaitAll, UInt32 dwMilliseconds);

		private HandleRef _handleReference;
		public WaitHelper(dynamic objectToWait, IntPtr? handle = null)
		{
			if (handle == null || handle == IntPtr.Zero)
				handle = objectToWait.Handle;

			_handleReference = new HandleRef(objectToWait, handle.Value);
		}

		public IntPtr Handle { get { return _handleReference.Handle; } }

		public bool Wait(Int32 milliSeconds = INFINITE)
		{
			return WAIT_TIMEOUT != WaitForSingleObject(_handleReference.Handle, (UInt32)milliSeconds);
		}

		public static bool WaitForAll(dynamic[] objectToWait, Int32 milliSeconds = INFINITE)
		{
			int nCount = objectToWait.Length;
			IntPtr[] handles = new IntPtr[nCount];
			WaitHelper[] waitHelpers = new WaitHelper[nCount];
			for (int i = 0; i < nCount; i++)
			{
				waitHelpers[i] = new WaitHelper(objectToWait[i]);
				handles[i] = waitHelpers[i].Handle;
			}

			int result = WaitForMultipleObjects(nCount, handles, 1, (UInt32)milliSeconds);
			return result >= WAIT_OBJECT_0 && result < (WAIT_OBJECT_0 + nCount);
		}

		public static int WaitForAny(dynamic[] objectToWait, Int32 milliSeconds = INFINITE)
		{
			int nCount = objectToWait.Length;
			IntPtr[] handles = new IntPtr[nCount];
			WaitHelper[] waitHelpers = new WaitHelper[nCount];
			for (int i = 0; i < nCount; i++)
			{
				waitHelpers[i] = new WaitHelper(objectToWait[i]);
				handles[i] = waitHelpers[i].Handle;
			}

			int result = WaitForMultipleObjects(nCount, handles, 0, (UInt32)milliSeconds);
			return result >= WAIT_OBJECT_0 && result < (WAIT_OBJECT_0 + nCount) ? result - 1 : -1;
		}
	}

	internal class BlockingThread
	{
		private ThreadStart _threadWorker;
		public Thread Thread { get; private set; }
		public IntPtr Handle;


		private ManualResetEvent taskRunningSem;
		private ManualResetEvent taskStartSem;

		private Exception _threadEx;

		public BlockingThread(ThreadStart threadStart)
		{
			_threadWorker = threadStart;
			Thread = new Thread(RunThread);

			// _thread.Priority = ThreadPriority.BelowNormal;

			taskStartSem = new ManualResetEvent(false);
			taskRunningSem = new ManualResetEvent(false);
			Handle = taskRunningSem.Handle;
		}

		public void Start()
		{
			taskStartSem.Reset();
			Thread.Start();

			// Warten bis der thread losläuft
			new WaitHelper(taskStartSem).Wait();
		}

		public bool Join(Int32 milliSeconds = WaitHelper.INFINITE)
		{
			WaitHelper waitHelper = new WaitHelper(taskRunningSem);
			bool result = waitHelper.Wait(milliSeconds);

			if (_threadEx != null)
				throw _threadEx;

			return result;
		}

		public bool NativeJoin(Int32 milliSeconds = WaitHelper.INFINITE)
		{
			if (milliSeconds == WaitHelper.INFINITE)
			{
				Thread.Join();
				return Join();
			}

			return Thread.Join(new TimeSpan(0, 0, 0, 0, milliSeconds));
		}

		private void RunThread()
		{
			try
			{
				_threadEx = null;
				taskRunningSem.Reset();

				// in der Start Methode kanns jetzt weitergehen
				taskStartSem.Set();

				// Hier machen wir die Arbeit
				if (_threadWorker != null)
					_threadWorker();
			}
			catch (Exception ex)
			{
				_threadEx = ex;
			}
			taskRunningSem.Set();
		}

		public static bool WaitThread(ThreadStart threadMethod, Int32 milliSeconds = WaitHelper.INFINITE)
		{
			BlockingThread t = new BlockingThread(threadMethod);
			t.Start();
			return t.Join(milliSeconds);
		}

		public static Thread GetNativeThread(dynamic thread)
		{
			if (thread != null)
			{
				if (thread is Thread)
					return thread;
				if (thread is BlockingThread)
					return thread.Thread;
			}

			return null;
		}
	}
}
