﻿/*
 * Copyright (c) 2010-2018 by Daniel Ranft, Wolfgang Hardt, Alexander Strobel
 *   Giegerich & Partner GmbH, Dreieich, Germany
 * Licensed under the MIT License, see LICENSE file for details.
 *
 *
 * Sourcecode: https://www.giepa.de/produkte/gpg4o/gnupgwrapper/
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using GupGPGWrapper.Properties;

namespace GupGPGWrapper
{
	public class GnuPGWrapper14 : GnuPGWrapper
	{
		public GnuPGWrapper14()
		{
			VersionUsed = new Version(1, 4, 0);
			UsePinEntry = false;
		}
	}

	public class GnuPGWrapper20 : GnuPGWrapper
	{
		public GnuPGWrapper20(bool usePinEntry)
		{
			VersionUsed = new Version(2, 0, 0);
			UsePinEntry = usePinEntry;
		}
	}

	public class GnuPGWrapper21 : GnuPGWrapper
	{
		public GnuPGWrapper21(bool usePinEntry)
		{
			VersionUsed = new Version(2, 1, 0);
			UsePinEntry = usePinEntry;
		}

		protected override void BuildCommand(StringBuilder commandAndParams, StringBuilder options)
		{
			base.BuildCommand(commandAndParams, options);

			if (PassphraseIsRequiredByCommand && !UsePinEntry)
				options.Append("--pinentry-mode=loopback ");
		}
	}

	public interface IGnuPGBlocking
	{
		void BlockingStarted();
		void BlockingStoppedTimeout();
		void BlockingStoppedNormal();
	}

	public interface IGnuPGBlockingWrapper : IDisposable
	{
		void StopBlockingGnuPG();
		bool StartBlockingGnuPG(int millisecondsTimeout, IGnuPGBlocking iBlockStop = null);
	}

	public class GnuPGBlockingWrapper : GnuPGWrapper, IGnuPGBlockingWrapper
	{
		public static readonly Version FixVersion = new Version(0, 1, 0);

		private IGnuPGBlocking _blockingInterface;

		public GnuPGBlockingWrapper()
		{
			VersionUsed = FixVersion;
			UsePinEntry = false;
			_blockingInterface = null;
		}

		public override void Dispose()
		{
			StopBlockingGnuPG();
		}

		/// <summary>
		/// Gets whether the member for <see cref="IGnuPGBlocking"/> has been set and therefore should be used to start/stop the blocking.
		/// </summary>
		private bool DoBlockingCallBack
		{
			get { return _blockingInterface != null; }
		}

		public void StopBlockingGnuPG()
		{
			if (BlockingGnuPG)
				InterruptExecute();
		}

		public bool StartBlockingGnuPG(int millisecondsTimeout, IGnuPGBlocking blockStopInterface = null)
		{
			StopBlockingGnuPG();

			_blockingInterface = blockStopInterface;

			return StartNoOperationTask(this, millisecondsTimeout);
		}

		/// <summary>
		/// Start a NOP which blocks other calls to GnuPG to prevent other threads from starting another GnuPG operation at the same time.
		/// </summary>
		/// <param name="wrapper"></param>
		/// <param name="millisecondsTimeout">In Milliseconds</param>
		/// <returns></returns>
		private static bool StartNoOperationTask(GnuPGBlockingWrapper wrapper, int millisecondsTimeout)
		{
			if (!wrapper.UseExecuteSynchronisation)
				return false;

			ManualResetEvent taskRunningSem = new ManualResetEvent(false);
			ManualResetEvent taskCancelSem = new ManualResetEvent(false);

			WaitHandle[] waitAnyEvent = new WaitHandle[] { InterruptWrapperSem, taskCancelSem };

			Task t = new Task(
				() =>
				{
					int idx = -1;
					lock (LockObjectExec)
					{
						try
						{
							taskRunningSem.Set();

							CurrentExecutionWrapper = wrapper;

							if (wrapper.DoBlockingCallBack)
								wrapper._blockingInterface.BlockingStarted();

							idx = WaitHandle.WaitAny(waitAnyEvent, millisecondsTimeout);
						}
						finally
						{
							CurrentExecutionWrapper = null;
							if (wrapper.DoBlockingCallBack)
							{
								if (idx == 0) // Index 0, Operation has successfully ended.
									wrapper._blockingInterface.BlockingStoppedNormal();
								else
									wrapper._blockingInterface.BlockingStoppedTimeout();
							}
						}
					}
				}
			);

			try
			{
				t.Start();

				// Don't wait for successful Task start, as we don't want to block by ourselves.
				if (taskRunningSem.WaitOne(millisecondsTimeout < 0 ? 3000 : millisecondsTimeout))
					return true;
			}
			catch { }

			taskCancelSem.Set();
			return false;
		}

		public override void Execute(out string outputString, int codepage = 0, params string[] inputStrings)
		{
			outputString = "No excution for blocking Wrapper!";
			throw new GnuPGCommandException(outputString);
		}

		public override void Execute(out string outputString, byte[] fileContent, int codepage = 0, params string[] inputStrings)
		{
			outputString = "No excution for blocking Wrapper!";
			throw new GnuPGCommandException(outputString);
		}
	}

	public class GnuPGWrapper : IDisposable
	{
		#region WinAPI Funktionen

		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern bool IsWindowVisible(IntPtr hWnd);

		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern int GetWindowText(IntPtr hWnd, StringBuilder lpString, int nMaxCount);

		private delegate bool EnumThreadDelegate(IntPtr hWnd, IntPtr lParam);
		[DllImport("user32.dll")]
		private static extern bool EnumThreadWindows(int dwThreadId, EnumThreadDelegate lpfn, IntPtr lParam);

		#endregion

		#region ECC curves name definitions
		/// <summary>
		/// Curve to GnuPG "Key-Curve:" Name resolution (for the unattended key generation)
		/// The curves are sorted by their trustworthy. The topmost curve (Keys[0]) is the most trustworthy.
		/// </summary>
		public static readonly Dictionary<string, string> GnuPG_Keygen_KeyCurves = new Dictionary<string, string>
		{
			{
				"Curve 25519",
				"Ed25519"
			},
			{
				"Brainpool P-512",
				"brainpoolP512r1"
			},
			{
				"Brainpool P-384",
				"brainpoolP384r1"
			},
			{
				"Brainpool P-256",
				"brainpoolP256r1"
			},
			{
				"NIST P-521",
				"nistp521"
			},
			{
				"NIST P-384",
				"nistp384"
			},
			{
				"NIST P-256",
				"nistp256"
			}
		};

		/// <summary>
		/// Curve to GnuPG "Key-Curve:" Name resolution (for the unattended key generation)
		/// The curves are sorted by their trustworthy. The topmost curve (Keys[0]) is the most trustworthy.
		/// </summary>
		public static readonly Dictionary<string, string> GnuPG_AddKey_KeyCurves = new Dictionary<string, string>
		{
			{
				"Curve 25519",
				"Curve25519"
			},
			{
				"Brainpool P-512",
				"brainpoolP512r1"
			},
			{
				"Brainpool P-384",
				"brainpoolP384r1"
			},
			{
				"Brainpool P-256",
				"brainpoolP256r1"
			},
			{
				"NIST P-521",
				"NIST P-521"
			},
			{
				"NIST P-384",
				"NIST P-384"
			},
			{
				"NIST P-256",
				"NIST P-256"
			}
		};
		#endregion

		private static object _lockObjectData = new object();

		#region Synchronisation, Blocking and Interrupt of Execution

		// Global signal used to interrupt GnuPG operations
		protected static readonly ManualResetEvent InterruptWrapperSem = new ManualResetEvent(false);

		// Sync objects
		private static object _lockObjectCurrentExecutionWrapper = new object();
		protected static readonly object LockObjectExec = new object();

		private static int _recursiveWrapperCnt;
		// Wrapper that is currently run
		private static GnuPGWrapper _currentExecutionWrapper;
		protected static GnuPGWrapper CurrentExecutionWrapper
		{
			set
			{
				lock (_lockObjectCurrentExecutionWrapper)
				{
					// Don't allow recursive Wrappers, only top level Wrappers are of importance
					if (value == null)
					{
						if (_recursiveWrapperCnt > 0)
							_recursiveWrapperCnt--;
					}
					else
					{
						_recursiveWrapperCnt++;
					}

					if (_recursiveWrapperCnt > 1)
						return;

					if (_currentExecutionWrapper != null)
						_currentExecutionWrapper.BlockingGnuPG = false;

					_currentExecutionWrapper = value;

					if (_currentExecutionWrapper != null)
					{
						_currentExecutionWrapper.BlockingGnuPG = true;
						InterruptWrapperSem.Reset();
					}
				}
			}
		}

		public bool BlockingGnuPG
		{
			get;
			private set;
		}

		/// <summary>
		/// Marks the GnuPG wrapper as interrupted
		/// </summary>
		public bool UserInterrupted
		{
			get;
			private set;
		}

		/// <summary>
		/// Sets the signal to interrupt the current Wrapper, if exists.
		/// </summary>
		/// <returns></returns>
		public static bool InterruptExecute()
		{
			lock (_lockObjectCurrentExecutionWrapper)
			{
				if (_currentExecutionWrapper == null || !_currentExecutionWrapper.BlockingGnuPG)
					return false;

				InterruptWrapperSem.Set();
				return true;
			}
		}
		#endregion

		private Process _process;

		private dynamic _outputThread;
		private dynamic _errorThread;
		private dynamic _inputThread;

		private Encoding _usedEncoding;
		private byte[] _inputFileContent;
		private string[] _inputStrings;

		public bool UseBlockingkWait;
		public bool UseThreadedWriting;
		public bool UseExecuteSynchronisation;

		public bool SuppressTimeoutMsg;

		#region Properties

		public delegate Process DelegateStartAgent(string homeDir);
		public delegate bool DelegateStopAgent(string homeDir, Process pinEntryProc);

		private DelegateStartAgent _startAgentDelegate;
		/// <summary>
		/// Delegate for starting the agent process
		/// </summary>
		public DelegateStartAgent StartAgentDelegate
		{
			set { _startAgentDelegate = value; }
			private get { return _startAgentDelegate; }
		}

		private DelegateStopAgent _stopAgentDelegate;
		/// <summary>
		/// Delegate for stopping the agent process
		/// </summary>
		public DelegateStopAgent StopAgentDelegate
		{
			set { _stopAgentDelegate = value; }
			private get { return _stopAgentDelegate; }
		}

		/// <summary>
		/// Sets the language, GnuPG 2 is using in its dialogs. It will be set as enviroment variable in ExecuteCommand()
		/// </summary>
		[DefaultValue("")]
		public static string GpgLanguage { private get; set; }

		/// <summary>
		/// Contains the last error messages and / or the log.
		/// </summary>
		/// <remarks>
		/// The data in this string don't need to be error messages
		/// </remarks>
		public string LastError { get; private set; }

		/// <summary>
		/// Returns the output of the last GnuPG process
		/// </summary>
		public string Output { get; private set; }

		/// <summary>
		/// Returns the status message of the last operation from <see cref="StatusFile"/> as string
		/// </summary>
		public string LastStatusString
		{
			get
			{
				return File.Exists(StatusFile) ? File.ReadAllText(StatusFile) : null;
			}
		}

		/// <summary>
		/// Gets or sets the path to the status file, which should be used. Defaults to string.Empty
		/// </summary>
		public string StatusFile { private get; set; }

		/// <summary>
		/// Get or set the character set which should be used (--charset)
		/// </summary>
		public string Charset { get; set; }

		[DefaultValue(Commands.SignAndEncrypt)]
		public Commands Command { private get; set; }

		public bool AsciiArmor { private get; set; }

		/// <summary>
		/// The list of Recipients email addresses or key-IDs - needed when <see cref="Command">Command</see> is one of the encryption commands
		/// </summary>
		public IEnumerable<string> Recipients { private get; set; }

		/// <summary>
		/// The originators email address or key-ID
		/// Recommended when <see cref="Command">Command</see> is one of the signing commands
		/// </summary>
		public string Originator { private get; set; }

		/// <summary>
		/// An additional recipient for encryption commands, normally this is your own key, so you can read your sent mails
		/// </summary>
		public string EncryptTo { private get; set; }

		/// <summary>
		/// If true, GnuPG allows usage of unsigned and/or untrusted keys. Defaults to false.
		/// </summary>
		public bool AlwaysTrust { private get; set; }

		public bool Batch { private get; set; }

		/// <summary>
		/// Don't emit GnuPG version
		/// </summary>
		public bool HideGnuPGVersion { private get; set; }

		/// <summary>
		/// The origin filename
		/// </summary>
		public string SetFileName { get; set; }

		private string _passphrase;
		public string Passphrase
		{
			private get { return _passphrase; }
			set
			{
				_passphrase = value;
				PassphraseFileDescriptor = string.IsNullOrEmpty(_passphrase) ? null : "0"; // stdin
			}
		}

		/// <summary>
		/// Path to the directory, where gpg.exe is located
		/// </summary>
		public string BinDirectory { private get; set; }

		/// <summary>
		/// Filename of the gpg[,2..n].exe
		/// </summary>
		public string GPGExecutable { private get; set; }

		/// <summary>
		/// File descriptor for entering passphrase - defaults to 0 (standard input).
		/// </summary>
		public string PassphraseFileDescriptor { private get; set; }

		/// <summary>
		/// URI of the keyserver to use
		/// </summary>
		public string KeyServer { private get; set; }

		/// <summary>
		/// Sets the Keys that should be affected by the operation
		/// Depending on the command, that can be be one or more IDs (Fingerprint, Long- or Short-ID) or a search string.
		/// </summary>
		public string KeyFilter { private get; set; }

		/// <summary>
		/// Sets the KeyIDs that should be affected by the operation
		/// Must be be one or more IDs (Fingerprint, Long- or Short-ID)
		/// </summary>
		public List<string> KeyFilterIDs { private get; set; }

		/// <summary>
		/// One based index of the User-ID that should be used in the next GnuPG action.
		/// Assign "0" to unselect all User-ID
		/// </summary>
		public int UserIDIndex { private get; set; }

		/// <summary>
		/// The (first) file, that should be processed by GnuPG.
		/// This filename will be passed to GnuPG as argument, not as byte[]
		/// </summary>
		public string FileName { private get; set; }

		/// <summary>
		/// The second file, that should be processed by GnuPG. This is needed for VerifyFile
		/// This filename will be passed to GnuPG as argument, not as byte[]
		/// </summary>
		public string SecondFileName { private get; set; }

		/// <summary>
		/// Exit-Code of GnuPG
		/// </summary>
		public int Exitcode { get; private set; }

		/// <summary>
		/// If true, --ignore-mdc-error flag is set during decryption proccess
		/// </summary>
		public bool IgnoreMdcError { private get; set; }

		[DefaultValue(VerboseLevel.NoVerbose)]
		public VerboseLevel Verbose { private get; set; }

		/// <summary>
		/// Timeout for GnuPG process, in milliseconds.
		/// For a better understanding of this property see the <see cref="WaitForProcessExit">WaitForProcessExit</see> method.
		/// </summary>
		[DefaultValue(15000)] // 15 seconds
		public int ProcessTimeOutMilliseconds { private get; set; }

		/// <summary>
		/// Sets the comment, which should be placed in the header of an ascii-armored message or file
		/// </summary>
		/// <remarks>Not used if <see cref="HideGnuPGVersion"/> is set to true</remarks>
		public string Comment { private get; set; }

		/// <summary>
		/// The file, where the output should be written to
		/// </summary>
		public string OutputFile { private get; set; }

		/// <summary>
		/// Alternate HomeDirectory (Keyrings, ...)
		/// </summary>
		public string HomeDir { private get; set; }

		/// <summary>
		/// Sets/Gets if Pinentry should be used
		/// </summary>
		/// <remarks>
		/// gpg v2.1 and later: If set to 'false', pinentry-mode=loopback will be appended to the gpg command line,
		/// when selected command requires a passphrase.
		/// </remarks>
		public bool UsePinEntry;

		/// <summary>
		/// Gets if selected command required a passphrase after the command has been executed
		/// </summary>
		/// <remarks>
		/// Property is used for adding 'pinentry-mode=loopback' to the corresponding gpg command line. 
		/// (In combination with <see cref="UsePinEntry"/> Property)
		/// </remarks>
		/// <!-- Property renamed, previous name was 'PassphraseNeeded'. (2019-01-04) -->
		protected bool PassphraseIsRequiredByCommand { get; private set; }

		/// <summary>
		/// Version of instantiated GnuPG(Wrapper)
		/// </summary>
		protected Version VersionUsed;

		/// <summary>
		/// Real version number of the used GnuPG
		/// </summary>
		public string VersionInfo { private get; set; }

		/// <summary>
		/// Copy GnuPG configuration file to new homedir if necessary
		/// </summary>
		/// <remarks><see cref="_copiedGpgConfFilename"/> for full filename</remarks>
		public bool CopyGpgConfToNewHomeDirIsActive { private get; set; }

		/// <summary>
		/// Full filename of gpg.conf-File copied when using another HomeDir
		/// Is deleted in Dispose()
		/// </summary>
		private string _copiedGpgConfFilename;

		/// <summary>
		/// Don't import missing keys for encryption
		/// </summary>
		public bool NoAutoKeyLocate { private get; set; }

		/// <summary>
		/// Don't import missing keys when verifying
		/// </summary>
		public bool NoAutoKeyRetrieve { private get; set; }

		/// <summary>
		/// When true GnuPG calls are made with "--force-mdc" option to activate Modification Detection Code
		/// </summary>
		public bool EncryptWithMdc { private get; set; }

		#endregion

		#region All the things needed for logging

		//////////////////////////////////////////
		//	The used log levels in GupLogger
		private const int cLogLevelLogStack = 7;
		private const int cLogLevelLogFunc = 6;
		private const int cLogLevelInfo = 5;
		private const int cLogLevelDebug = 4;
		private const int cLogLevelError = 3;
		private const int cLogLevelFatal = 2;
		private const int cLogLevelAlways = 1;

		/// <summary>
		/// delegate for logging - just set your function as LoggingMethod
		/// and you can let GnuPGWrapper write to your application logfile
		/// </summary>
		public delegate void LoggingDelegate(string debugString, int logLevel, int options = 0);

		private LoggingDelegate _loggingMethod;
		public LoggingDelegate LoggingMethod
		{
			private get { return _loggingMethod ?? LoggingDummyMethod; }
			set { _loggingMethod = value; }
		}

		/// <summary>
		/// A dummy method, which is used, if LogginMethod hasn't been set
		/// </summary>
		private static void LoggingDummyMethod(string s, int logLevel, int options) { }

		#endregion

		protected GnuPGWrapper()
		{
			OutputFile = null;
			ProcessTimeOutMilliseconds = 15000;
			Command = Commands.SignAndEncrypt;
			Batch = true;
			UsePinEntry = true;
			PassphraseIsRequiredByCommand = false;
			_copiedGpgConfFilename = string.Empty;
			StartAgentDelegate = null;
			StopAgentDelegate = null;
		}

		~GnuPGWrapper()
		{
			Dispose();
		}

		private static void DeleteFile(string fileToDelete)
		{
			if (string.IsNullOrEmpty(fileToDelete) || !File.Exists(fileToDelete))
				return;

			try
			{
				File.Delete(fileToDelete);
			}
			catch { }
		}

		public virtual void Dispose()
		{
			if (Command != Commands.AddSubKey)
			{
				DeleteFile(StatusFile);
				StatusFile = "";
			}

			DeleteFile(_copiedGpgConfFilename);
			_copiedGpgConfFilename = "";
		}

		/// <summary>
		/// Returns the HomeDir of the current GnuPG
		/// </summary>
		/// <returns></returns>
		private string GetHomeDirFromVersionInfo()
		{
			string homeDir = string.Empty;
			try
			{
				Match m = Regex.Match(VersionInfo, @"^home: (.*?)$", RegexOptions.Multiline | RegexOptions.IgnoreCase);
				homeDir = m.Groups[1].Value;
				if (!string.IsNullOrEmpty(homeDir))
				{
					homeDir = homeDir.Trim(' ', '\r', '\n', '\t');
					homeDir = homeDir.Replace("/", "\\");
				}
			}
			catch
			{
			}

			return homeDir;
		}

		/// <summary>
		/// Copies the gpg.conf from the current to the new HomeDir if neccessary
		/// </summary>
		/// <param name="newHomeDir"></param>
		private void CopyGpgConfToHomeDir(string newHomeDir)
		{
			try
			{
				if (!CopyGpgConfToNewHomeDirIsActive)
					return;

				if (string.IsNullOrEmpty(newHomeDir))
					return;

				if (!Directory.Exists(newHomeDir))
					return;

				string gpgConfFile = Path.Combine(newHomeDir, "gpg.conf");
				if (File.Exists(gpgConfFile))
					return;

				string curHomeDir = GetHomeDirFromVersionInfo();
				if (string.IsNullOrEmpty(curHomeDir))
				{
					LoggingMethod("GnuPG: No HomeDir !!!", cLogLevelDebug);
					return;
				}

				string gpgConfFileOrg = Path.Combine(curHomeDir, "gpg.conf");
				if (!File.Exists(gpgConfFileOrg))
					return;

				try
				{
					File.Copy(gpgConfFileOrg, gpgConfFile);
					_copiedGpgConfFilename = gpgConfFile;
				}
				catch
				{
					LoggingMethod("GnuPG: gpg.conf not copied to new homedir !!!", cLogLevelDebug);
				}
			}
			catch (Exception ex)
			{
				LoggingMethod("GnuPG: CopyConfFile Exception:: \r\n" + ex.Message, cLogLevelDebug);
			}
		}

		/// <summary>
		/// Creates the neccessary command line arguments for a GnuPG command
		/// </summary>
		/// <param name="options"></param>
		/// <remarks>Can be overriden in derived classes</remarks>
		protected virtual void BuildOptions(StringBuilder options)
		{
			options.Append("--yes ");

			if (!string.IsNullOrEmpty(HomeDir))
			{
				// Bug 1377: Das HomeDir beim Ermitteln der GnuPG Version nicht mehr übergeben (führt zu einer Exception)
				if (Command != Commands.Version)
				{
					Directory.CreateDirectory(HomeDir);
					options.AppendFormat("--homedir \"{0}\" ", HomeDir);
					CopyGpgConfToHomeDir(HomeDir);
				}
			}

			if (!string.IsNullOrEmpty(StatusFile))
				options.AppendFormat("--status-file \"{0}\" ", StatusFile);

			if (!string.IsNullOrEmpty(KeyServer))
				options.AppendFormat("--keyserver {0} ", KeyServer);

			if (Batch)
				options.Append("--batch ");

			if (!string.IsNullOrEmpty(Charset))
				options.AppendFormat("--charset {0} ", Charset);

			if (AlwaysTrust)
				options.Append("--always-trust ");

			if (IgnoreMdcError)
				options.Append("--ignore-mdc-error ");

			if (EncryptWithMdc)
				options.Append("--force-mdc ");

			if (!string.IsNullOrEmpty(SetFileName))
				options.AppendFormat("--utf8-strings --set-filename \"{0}\" ", Encoding.Default.GetString(Encoding.UTF8.GetBytes(SetFileName)));

			if (Command == Commands.ListPublicKeys || Command == Commands.ListSecretKeys ||
				Command == Commands.ListSignatures || Command == Commands.SearchKeys)
			{
				options.Append("--with-colons --fixed-list-mode ");
			}
			else if (Command == Commands.EncryptFile || Command == Commands.SignFile || Command == Commands.SignAndEncryptFile)
			{
				// Bug 1402 - Wenn in der Config-Datei utf8-strings angegeben wurde, entstehen hässliche Umlaute
				options.Append("--no-utf8-strings ");
			}

			if (AsciiArmor)
			{
				options.Append("--armor ");
				// Bug 3563, Kommentare nur, wenn HideGnuPGHeaders nicht aktiviert ist
				if (!HideGnuPGVersion && !string.IsNullOrEmpty(Comment))
					options.AppendFormat("--comment \"{0}\" ", Comment.Replace("\r", "").Replace("\n", "")); // Comment without linefeeds!!!
			}

			if (!string.IsNullOrWhiteSpace(Originator))
			{
				string tmp = (Command == Commands.SignKey) ? "-u" : "--default-key";
				options.AppendFormat("{0} {1} ", tmp, Originator);
			}

			if (!string.IsNullOrWhiteSpace(OutputFile))
				options.AppendFormat("--output \"{0}\" ", OutputFile);

			switch (Verbose)
			{
				case VerboseLevel.NoVerbose:
					options.Append("--no-verbose ");
					break;
				case VerboseLevel.Verbose:
					options.Append("--verbose ");
					break;
				case VerboseLevel.VerboseVerbose:
					options.Append("--verbose --verbose ");
					break;
			}

			if (HideGnuPGVersion)
			{
				// Bug 3563: Suppress GnuPG version only (don't suppress comments anymore)
				options.Append("--no-emit-version  ");
			}

			if (!string.IsNullOrEmpty(EncryptTo))
				options.AppendFormat("--encrypt-to {0} ", EncryptTo);

			if (Recipients != null && Recipients.Any())
			{
				foreach (string recipient in Recipients)
					options.AppendFormat("--recipient {0} ", recipient);
			}

			if (!string.IsNullOrEmpty(PassphraseFileDescriptor))
				options.AppendFormat("--passphrase-fd {0} ", PassphraseFileDescriptor); // always "0" for stdin

			if (NoAutoKeyLocate)
				options.Append("--no-auto-key-locate ");

			if (NoAutoKeyRetrieve)
				options.Append("--no-auto-key-retrieve ");
		}

		/// <summary>
		/// Appends a filter at the end of the command.
		/// A filter is for example a unique id (email address) that will be quoted or
		/// a list of key ids that will be append without any quotes separated by blanks
		/// </summary>
		/// <param name="commandAndParams"></param>
		protected void AddFilterToCommand(StringBuilder commandAndParams)
		{
			if (!string.IsNullOrEmpty(KeyFilter))
				commandAndParams.AppendFormat(" \"{0}\" ", KeyFilter);
			else if (KeyFilterIDs != null && KeyFilterIDs.Count > 0)
				commandAndParams.AppendFormat(" {0} ", string.Join(" ", KeyFilterIDs));
			// If there is no filter activ, than all results will be returned
		}

		/// <summary>
		/// Appends the file names afte the GnuPG command
		/// </summary>
		/// <param name="commandAndParams"></param>
		/// <param name="fileName">Primary file to process (i.e. detached signed file)</param>
		/// <param name="secondFileName">Secondary file to process (i.e. detached signature file)</param>
		protected void AddFilesToCommand(StringBuilder commandAndParams, string fileName, string secondFileName = "")
		{
			if (!string.IsNullOrEmpty(fileName))
				commandAndParams.AppendFormat("\"{0}\" ", fileName);
			if (!string.IsNullOrEmpty(secondFileName))
				commandAndParams.AppendFormat("\"{0}\" ", secondFileName);
		}

		/// <summary>
		/// Appends the file names after the GnuPG command
		/// </summary>
		/// <param name="commandAndParams"></param>
		protected void AddFilesToCommand(StringBuilder commandAndParams)
		{
			AddFilesToCommand(commandAndParams, FileName, SecondFileName);
		}

		/// <summary>
		/// Creates the neccessary command line arguments for a GnuPG command
		/// </summary>
		/// <param name="commandAndParams"></param>
		/// <param name="options"></param>
		/// <remarks>Can be overriden in derived classes</remarks>
		protected virtual void BuildCommand(StringBuilder commandAndParams, StringBuilder options)
		{
			bool recipientNeeded = false;

			#region Select and prepare command action

			switch (Command)
			{
				case Commands.Sign:
					commandAndParams.Append("--clearsign ");
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.SignFile:
					commandAndParams.Append("--detach-sign ");
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.Encrypt:
					commandAndParams.Append("--encrypt ");
					recipientNeeded = true;
					break;
				case Commands.EncryptFile:
					commandAndParams.Append("--encrypt ");
					recipientNeeded = true;
					break;
				case Commands.SignAndEncrypt:
					commandAndParams.Append("--sign --encrypt ");
					recipientNeeded = true;
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.SignAndEncryptFile:
					commandAndParams.Append("--sign --encrypt ");
					recipientNeeded = true;
					break;
				case Commands.Decrypt:
					commandAndParams.Append("--decrypt ");
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.DecryptFile:
					commandAndParams.Append("--decrypt ");
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.Verify:
					commandAndParams.Append("--verify ");
					AddFilesToCommand(commandAndParams, FileName);	// signature file 
					commandAndParams.Append(" - ");					// data to verify will be streamed to GnuPG stdin  
					break;
				case Commands.VerifyFile:
					commandAndParams.AppendFormat("--verify ");
					AddFilesToCommand(commandAndParams);
					break;
				case Commands.GenerateKey:
					commandAndParams.Append("--gen-key ");
					break;
				case Commands.ExportPublicKey:
					commandAndParams.Append("--export ");
					AddFilterToCommand(commandAndParams);
					Batch = this is GnuPGWrapper21 ? false : Batch;
					break;
				case Commands.ExportSecretKey:
					commandAndParams.Append("--export-secret-keys ");
					AddFilterToCommand(commandAndParams);
					PassphraseIsRequiredByCommand = true;
					Batch = this is GnuPGWrapper21 ? false : Batch;
					break;
				case Commands.ImportKeyFile:
					commandAndParams.AppendFormat("--import ");
					AddFilesToCommand(commandAndParams);
					if (Passphrase != null)
						PassphraseIsRequiredByCommand = true;
					break;
				case Commands.DeleteKey:
					commandAndParams.AppendFormat("--delete-secret-and-public-key ");
					AddFilterToCommand(commandAndParams);
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.ListPackets:
					commandAndParams.Append("--list-packets ");
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.ListEncryptionKeys:
					options.Append("--list-only ");
					commandAndParams.Append("--decrypt ");
					break;
				case Commands.ListPacketsOnly:
					options.Append("--list-only ");
					commandAndParams.Append("--list-packets ");
					break;
				case Commands.ListPublicKeys:
					options.Append("--fingerprint --fingerprint ");
					commandAndParams.Append("--list-public-keys ");
					AddFilterToCommand(commandAndParams);
					break;
				case Commands.ListSecretKeys:
					options.Append("--fingerprint --fingerprint ");
					commandAndParams.Append("--list-secret-keys ");
					AddFilterToCommand(commandAndParams);
					break;
				case Commands.ListSignatures:
					options.Append("--fingerprint --fingerprint ");
					commandAndParams.Append("--list-sigs ");
					AddFilterToCommand(commandAndParams);
					break;
				case Commands.Version:
					commandAndParams.Append("--version ");
					break;
				case Commands.SignKey:
					options.Append("--command-fd 0 --ask-cert-level ");
					commandAndParams.AppendFormat("--sign-key \"{0}\" ", KeyFilter);
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.LocalSignKey:
					options.Append("--command-fd 0 --ask-cert-level ");
					commandAndParams.AppendFormat("--lsign-key \"{0}\" ", KeyFilter);
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.TrustKey:
					options.Append("--command-fd 0 ");
					commandAndParams.AppendFormat("--edit-key {0} trust ", KeyFilter);
					Batch = this is GnuPGWrapper21 ? false : Batch;
					break;
				case Commands.EnableKey:
					options.Append("--command-fd 0 ");
					commandAndParams.AppendFormat("--edit-key {0} enable ", KeyFilter);
					Batch = this is GnuPGWrapper21 ? false : Batch;
					break;
				case Commands.DisableKey:
					options.Append("--command-fd 0 ");
					commandAndParams.AppendFormat("--edit-key {0} disable ", KeyFilter);
					Batch = this is GnuPGWrapper21 ? false : Batch;
					break;
				case Commands.ChangePassphrase:
					options.Append("--command-fd 0 ");
					commandAndParams.AppendFormat("--edit-key {0} passwd ", KeyFilter);
					Batch = this is GnuPGWrapper21 ? false : Batch;
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.SetExpireDate:
					options.Append("--command-fd 0 ");
					commandAndParams.AppendFormat("--edit-key {0} expire", KeyFilter);
					Batch = this is GnuPGWrapper21 ? false : Batch;
					break;
				case Commands.EditKey:
					options.Append("--command-fd 0 ");
					commandAndParams.AppendFormat("--edit-key {0} ", KeyFilter);
					Batch = this is GnuPGWrapper21 ? false : Batch;
					break;
				case Commands.AddUserId:
					options.Append("--command-fd 0 ");
					options.Append("--allow-freeform-uid ");
					commandAndParams.AppendFormat("--edit-key {0} adduid", KeyFilter);
					Batch = this is GnuPGWrapper21 ? false : Batch;
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.SetPrimaryUserId:
					options.Append("--command-fd 0 ");
					commandAndParams.AppendFormat("--edit-key {0} \"uid {1}\" primary", KeyFilter, UserIDIndex);
					Batch = this is GnuPGWrapper21 ? false : Batch;
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.SearchKeys:
					options.Append("--command-fd 0 ");
					commandAndParams.AppendFormat("--search-keys {0} ", KeyFilter);
					break;
				case Commands.ReceiveKeys:
					commandAndParams.AppendFormat("--recv-keys {0} ", KeyFilter);
					break;
				case Commands.SendKeys:
					commandAndParams.AppendFormat("--send-keys {0} ", KeyFilter);
					break;
				case Commands.CreateRevocationCertificate:
					options.Append("--command-fd 0 ");
					commandAndParams.AppendFormat("--gen-revoke {0} ", KeyFilter);
					PassphraseIsRequiredByCommand = true;
					break;
				case Commands.ImportOwnertrust:
					commandAndParams.Append("--import-ownertrust ");
					break;
				case Commands.ExportOwnertrust:
					commandAndParams.Append("--export-ownertrust ");
					break;
				case Commands.ImportRevocationCertificate:
					commandAndParams.Append("--import ");
					break;
				case Commands.ListConfig:
					options.Append("--with-colons ");
					commandAndParams.Append("--list-config ");
					break;
				case Commands.ReSignWeakSignedKey:
					options.Append("--command-fd 0 ");
					options.Append("--allow-weak-digest ");
					options.Append("--cert-digest-algo SHA512 ");
					commandAndParams.AppendFormat("--edit-key {0} expire", KeyFilter);
					break;
				case Commands.AddSubKey:
					options.Append("--expert ");
					options.Append("--command-fd 0 ");
					commandAndParams.AppendFormat("--edit-key {0} addkey", KeyFilter);
					break;
				default:
					throw new GnuPGException("This command isn't implemented");
			}
			#endregion

			if (recipientNeeded && (Recipients == null || !Recipients.Any()))
				throw new GnuPGException("Cannot encrypt with empty Recipients");

		}

		private string BuildArguments()
		{
			StringBuilder mainOptions = new StringBuilder(255);
			StringBuilder additionalOptions = new StringBuilder(255);
			StringBuilder commandAndParams = new StringBuilder(30);

			BuildCommand(commandAndParams, additionalOptions);
			BuildOptions(mainOptions);

			mainOptions.Append(additionalOptions);

			return mainOptions.ToString().Trim() + " " + commandAndParams.ToString().Trim();
		}

		/// <summary>
		/// Determines the caracter sets name 
		/// </summary>
		/// <param name="codepage"></param>
		/// <returns></returns>
		/// <!-- Bug 1710 -->
		private Encoding GetEncodingAndSetCharset(int codepage)
		{
			// Bug 712: Charset setzen damit es über BuildArguments() in den Kommandozeilenstring für GnuPG kommt.
			Encoding gpgEncoding = Encoding.Default;
			if (codepage == 0)
				Charset = string.Empty;
			else
			{
				if (codepage == Encoding.UTF8.CodePage)
					gpgEncoding = new UTF8Encoding(false, false);	// Process data without BOM
				else
					gpgEncoding = Encoding.GetEncoding(codepage);

				Charset = gpgEncoding.WebName;		// z.B. iso-8859-x, koi8-x, windows-1251, ...
			}

			return gpgEncoding;
		}

		public virtual void Execute(out string outputString, int codepage = 0, params string[] inputStrings)
		{
			byte[] content = new byte[0];

			if (inputStrings.Length > 0)
			{
				// Bug 1805: Exception verhindern, wenn bspw. ein leerer Body verarbeitet werden soll.
				if (inputStrings[0] == null)
					inputStrings[0] = string.Empty;

				if (File.Exists(inputStrings[0]))
					content = File.ReadAllBytes(inputStrings[0]);
				else
					content = GetEncodingAndSetCharset(codepage).GetBytes(inputStrings[0]);
			}

			Execute(out outputString, content, codepage, inputStrings);
		}

		public virtual void Execute(out string outputString, byte[] fileContent, int codepage = 0, params string[] inputStrings)
		{
			// All parallel Wrappers must wait
			if (UseExecuteSynchronisation)
				Monitor.Enter(LockObjectExec);

			try
			{
				try
				{
					ProcessExecute(out outputString, fileContent, codepage, inputStrings);
				}
				catch (GnuPGAgentException)
				{
					// Retry when restarting the Agent
					ProcessExecute(out outputString, fileContent, codepage, inputStrings);
				}
			}
			finally
			{
				if (UseExecuteSynchronisation)
					Monitor.Exit(LockObjectExec);
			}
		}

		private void ProcessExecute(out string outputString, byte[] fileContent, int codepage = 0, params string[] inputStrings)
		{
			long startTick = Environment.TickCount;
			string fullPathToGpg = Path.Combine(BinDirectory, GPGExecutable);

			// Existenz der gpg.exe kontrollieren
			if (!new FileInfo(fullPathToGpg).Exists)
				throw new GnuPGException("GPGExecutable not found!");

			string arguments = BuildArguments();
			LoggingMethod("GnuPG: %GPGPATH%\\" + GPGExecutable + " " + arguments, cLogLevelDebug, 0x01 | 0x02);

			// Create startinfo object
			_process = new Process();
			_process.StartInfo.Arguments = arguments;
			_process.StartInfo.FileName = fullPathToGpg;
			_process.StartInfo.WorkingDirectory = BinDirectory;
			_process.StartInfo.CreateNoWindow = true;
			_process.StartInfo.UseShellExecute = false;

			// Redirect everything: 
			_process.StartInfo.RedirectStandardInput = true;	// to send the passphrase and, if needed, the data to process
			_process.StartInfo.RedirectStandardOutput = true;	// to get encrypted message
			_process.StartInfo.RedirectStandardError = true;	// to get error messages

			// Umgebungsvariable setzen, so dass der Pinentry mit den lokalisierten Texten kommt (24.01.2012 A.Strobel)
			string envVariableNameLanguage = "lang";
			if (this is GnuPGWrapper14)
				envVariableNameLanguage = "language";

			if (_process.StartInfo.EnvironmentVariables.ContainsKey(envVariableNameLanguage))
				_process.StartInfo.EnvironmentVariables.Remove(envVariableNameLanguage);
			_process.StartInfo.EnvironmentVariables.Add(envVariableNameLanguage, GpgLanguage);

			// DR: Adding codepage-support
			_usedEncoding = GetEncodingAndSetCharset(codepage);
			_process.StartInfo.StandardOutputEncoding = _usedEncoding;
			_process.StartInfo.StandardErrorEncoding = _usedEncoding;

			_process.Start();

			Output = string.Empty;
			LastError = string.Empty;

			_outputThread = UseBlockingkWait ? (object)new BlockingThread(ReadStandardOutput) : new Thread(ReadStandardOutput);
			_outputThread.Start();
			_errorThread = UseBlockingkWait ? (object)new BlockingThread(ReadStandardError) : new Thread(ReadStandardError);
			_errorThread.Start();

			// Die Daten an den wrapper binden 
			_inputFileContent = fileContent;
			_inputStrings = inputStrings;

			// Synchron oder asynchron
			WriteStandardInput();

			// Wait for the end of the process 
			Exitcode = WaitForProcessExit();

			// Check results and prepare Output
			outputString = Output;
			long endTick = Environment.TickCount;

			if (Exitcode == 0)
			{
				LoggingMethod("GnuPG: Exit Code: " + Exitcode + " (" + (endTick - startTick) + "ms)", cLogLevelDebug);

				/////////////////////////////////////////////////////////////////
				// Some Commands transmit the relevant data in error stream
				if (Command == Commands.SendKeys ||
					Command == Commands.Verify)
				{
					if (string.IsNullOrEmpty(outputString) && !string.IsNullOrEmpty(LastError))
						outputString = LastError;
				}
				/////////////////////////////////////////////////////////////////
			}
			else
			{
				if (string.IsNullOrEmpty(LastError))
					LastError = "Unknown GnuPG-Error with exitcode " + Exitcode;

				if (LastError.EndsWith("\r\n"))
					LastError = LastError.Substring(0, LastError.Length - 2);

				// Bug 3530: Bei Fehlern in "list-packets(-only)" nicht einfach vollständig in das Log rausschreiben
				string lastErrorToLog = LastError;
				if ((Command == Commands.ListPackets || Command == Commands.ListPacketsOnly)
					&& lastErrorToLog.Length > 500)
				{
					lastErrorToLog = lastErrorToLog.Substring(0, 499) + " ...";
				}

				LoggingMethod("GnuPG: Exit Code: " + Exitcode + " (" + (endTick - startTick) + "ms)" + " / " + lastErrorToLog, cLogLevelDebug);

				// With a GnuPG exit code unlike zero we generally throw an exception
				// throw new GnuPGCommandException(LastError, Output);				// Output -> Could contain decrypted Data
				// throw new GnuPGCommandException(LastError, LastStatusString);	// LastStatusString -> could contain the USERID's
				throw new GnuPGCommandException(LastError);	 						// no further information in exception message to guarantee privacy	
			}
		}

		#region input stream handling
		private void CopyDataToStream(byte[] data, Stream output)
		{
			MemoryStream input = new MemoryStream(data);

			// ReSharper disable once InconsistentNaming
			const int BUFFER_SIZE = 4096;
			byte[] bytes = new byte[BUFFER_SIZE];
			int i;
			while ((i = input.Read(bytes, 0, bytes.Length)) != 0)
			{
				output.Write(bytes, 0, i);
			}
			output.Flush();
		}

		private bool HaveDataForStdInput()
		{
			return !string.IsNullOrEmpty(Passphrase) ||
					(_inputFileContent != null && _inputFileContent.Length > 0) ||
					(_inputStrings != null && _inputStrings.Length > 1);
		}


		private bool IamTheInputThread()
		{
			Thread nativeThread = BlockingThread.GetNativeThread(_inputThread);
			return nativeThread != null && nativeThread.IsAlive && nativeThread.ManagedThreadId == Thread.CurrentThread.ManagedThreadId;
		}

		/// <summary>
		/// Writer thread for stdin gpg
		/// </summary>
		private void WriteStandardInput()
		{
			// Do i need it thread for writing data to stdin
			if (UseThreadedWriting			// ja, wenn aktiviert und 
				&& HaveDataForStdInput() 	// wenn Daten zum schreiben da sind
				&& !IamTheInputThread())	// und wenn ich selbst noch nicht in diesem Thread laufe
			{
				// Threaded writing
				_inputThread = UseBlockingkWait ? (object)new BlockingThread(WriteStandardInput) : new Thread(WriteStandardInput);
				_inputThread.Start();
				return;
			}

			try
			{
				// Bei RevoCert wird die Passphrase am Schluss eingegeben
				// if (!string.IsNullOrEmpty(Passphrase) && Command != Commands.CreateRevocationCertificate)
				if (!string.IsNullOrEmpty(Passphrase))
				{
					_process.StandardInput.WriteLine(Passphrase);
					_process.StandardInput.Flush();
				}

				if (_inputFileContent != null && _inputFileContent.Length > 0)
				{
					CopyDataToStream(_inputFileContent, _process.StandardInput.BaseStream);
				}

				if (_inputStrings != null && _inputStrings.Length > 1)
				{
					using (StreamWriter inputStream = new StreamWriter(_process.StandardInput.BaseStream, _usedEncoding))
					{
						inputStream.AutoFlush = _process.StandardInput.AutoFlush;
						inputStream.NewLine = _process.StandardInput.NewLine;

						// Send input text, if any
						for (int idx = 1; idx < _inputStrings.Length; idx++)
						{
							if (string.IsNullOrEmpty(_inputStrings[idx]))
								continue;

							if (string.Equals(Environment.NewLine, _inputStrings[idx]))
								inputStream.WriteLine();
							else if (_inputStrings[idx].EndsWith(Environment.NewLine))
								inputStream.Write(_inputStrings[idx]);
							else
								inputStream.WriteLine(_inputStrings[idx]);
							inputStream.Flush();
						}

						inputStream.Close();
					}
				}
				else
				{
					_process.StandardInput.Close();
				}
			}
			catch (ThreadAbortException ex)
			{
				lock (_lockObjectData)
				{
					if (ex.ExceptionState != null)
						LastError += "GnuPG-Error: (WriteStandardInput) " + ex.ExceptionState + "\r\n";
				}
			}
			catch
			{
				lock (_lockObjectData)
				{
					LastError += "GnuPG-Error: (WriteStandardInput) FATAL_ERROR\r\n";
				}
			}
		}

		#endregion

		#region output and error stream handling
		/// <summary>
		/// Reader thread for stdout
		/// </summary>
		private void ReadStandardOutput()
		{
			try
			{
				using (StreamReader outputStream = new StreamReader(_process.StandardOutput.BaseStream, _usedEncoding))
				{
					string s = outputStream.ReadToEnd();
					lock (_lockObjectData)
					{
						Output = s;
					}
				}
			}
			catch (ThreadAbortException ex)
			{
				lock (_lockObjectData)
				{
					if (ex.ExceptionState != null)
						Output = "GnuPG-Error: (ReadStandardOutput) " + ex.ExceptionState;
				}
			}
			catch
			{
				lock (_lockObjectData)
				{
					Output = "GnuPG-Error: (ReadStandardOutput) FATAL_ERROR";
				}
			}
		}

		/// <summary>
		/// Reader thread for stderr
		/// </summary>
		private void ReadStandardError()
		{
			try
			{
				string s = _process.StandardError.ReadToEnd();
				lock (_lockObjectData)
				{
					LastError = s;
				}
			}
			catch (ThreadAbortException ex)
			{
				lock (_lockObjectData)
				{
					if (ex.ExceptionState != null)
						LastError += "GnuPG-Error: (ReadStandardError) " + ex.ExceptionState + "\r\n";
				}
			}
			catch
			{
				lock (_lockObjectData)
				{
					LastError += "GnuPG-Error: (ReadStandardError) FATAL_ERROR\r\n";
				}
			}
		}
		#endregion

		#region gpg process controlling

		/// <summary>
		/// Wait for the end of the GPG[2] process
		/// </summary>
		/// <returns>Exit code of the process</returns>
		private int WaitForProcessExit()
		{
			bool bWaitForProccessEnd = true;
			while (bWaitForProccessEnd)
			{
				// Normal mit Timeout warten
				if (WaitProcEnd(_process, ProcessTimeOutMilliseconds, UseExecuteSynchronisation))
				{
					// Process exited before timeout...
					// Wait for the threads to complete reading Output/error (but use a timeout!)
					bWaitForProccessEnd = false;
					JoinAndAbortThread(_inputThread, ProcessTimeOutMilliseconds, true, "_inputThread timeout");
					JoinAndAbortThread(_outputThread, ProcessTimeOutMilliseconds / 2, true, "_outputThread timeout");
					JoinAndAbortThread(_errorThread, ProcessTimeOutMilliseconds / 2, true, "_errorThread timeout");
				}
				else
				{
					// Sollte der Dialog zur Eingabe der Pin/Passphrase offen sein, 
					// wird weiter gewartet. Dabei erfolgt ein Neustart des Timeouts, 
					// der immer wieder prüft, ob der Dialog noch offen ist.
					// ist er nicht mehr offen (warum auch immer) so kommt unten der Hinweis
					// dass der GPG-Prozess zu lange braucht und ob man abbrechen will.
					bool isPinentryVisible;
					Process pinentry = CheckForPinentry(out isPinentryVisible);
					bool pinentryFound = pinentry != null;

					// Der Nutzer kann/soll Eingaben vornehmen: Wir warten weiter.
					if (pinentryFound && isPinentryVisible)
						continue;

					// Frage nach mehr Zeit an Nutzer stellen, wenn es um Hintergrundprozesse geht
					bool askForMoreTime = !UserInterrupted && !pinentryFound && !isPinentryVisible && !SuppressTimeoutMsg;

					// Process timeout: PGP hung somewhere... kill it (as well as the threads!)
					bool restartAgent = false;
					if (askForMoreTime)
					{
						DialogResult res = MessageBox.Show(Resources.STR_PROCCESSTIMEOUT, Resources.STR_ERROR, MessageBoxButtons.YesNo, MessageBoxIcon.Question);
						restartAgent = res == DialogResult.No;
						if (!restartAgent)
							continue;		// User wants to wait for command to terminate
					}

					// Kill/Resart GnuPG processes
					bool agentRestarted = ResetGPGProcesses(_process, pinentry, restartAgent);

					////////////////
					// First, terminate the treads if they are still running...
					ThreadAbortSave(_inputThread, "_inputThread timeout");
					ThreadAbortSave(_errorThread, "_errorThread timeout");
					ThreadAbortSave(_outputThread, "_outputThread timeout");

					// Second, set the thread outputs to defined values
					Output = string.Empty;

					// Bug 3656: specialized LastError when process was cancelled by user
					if (UserInterrupted)
						LastError = "Process execution aborted by user!\r\n";
					else
						LastError = "Process timed out after " + ProcessTimeOutMilliseconds.ToString(CultureInfo.InvariantCulture) + " milliseconds\r\n";
					////////////////

					// Bug 3620: throw specific exception typ to retry the last GnuPG command
					if (agentRestarted)
					{
						LoggingMethod("After agent restart force a second try for the last GnuPG command!", cLogLevelInfo);
						throw new GnuPGAgentException("gpg-agent restarted. GnuPG command will be executed again!");
					}

					break;
				}
			}

			return _process.ExitCode;
		}

		/// <summary>
		/// Wait for process ends 
		/// </summary>
		/// <param name="process">process which will be wait for</param>
		/// <param name="timeout">time to wait for process ending </param>
		/// <param name="isInterruptable">Waiting for process end can be cancelled by user</param>
		/// <returns>false when the process still is running (timeout occured, when waiting for its end)</returns>
		private bool WaitProcEnd(Process process, int timeout, bool isInterruptable)
		{
			try
			{
				if (isInterruptable)
					CurrentExecutionWrapper = this;

				return WaitProcEndInterruptable(process, timeout, isInterruptable);
			}
			finally
			{
				if (isInterruptable)
					CurrentExecutionWrapper = null;
			}
		}

		/// <summary>
		/// Wait for process to end, but in a way that this can be interrupted
		/// </summary>
		/// <param name="process">Process to wait for</param>
		/// <param name="timeout">Timeout after which the function will return anyway.</param>
		/// <returns>false when the process still is running (timeout occured or user cancelled, when waiting for its end)</returns>
		/// <remarks>
		/// Values -1 and int.MaxValue for <paramref name="timeout"/> count as "infinite wait". 
		/// In these cases there won't be any timeout ever.
		/// </remarks>
		private bool WaitProcEndInterruptable(Process process, int timeout, bool handleInterrupt)
		{
			// Use Native Blocking
			WaitHelper proccessWaitHelper = null;
			WaitHelper interruptWrapperSemWaitHelper = null;
			if (UseBlockingkWait)
			{
				try { proccessWaitHelper = new WaitHelper(process); }
				catch { }

				try { interruptWrapperSemWaitHelper = handleInterrupt ? new WaitHelper(InterruptWrapperSem) : null; }
				catch { }
			}

			bool haveTimeout = timeout > -1 && timeout != int.MaxValue;
			DateTime dtWaitUntil = DateTime.Now;
			if (haveTimeout)
				dtWaitUntil = dtWaitUntil.AddMilliseconds(timeout);

			while (true)
			{
				if (proccessWaitHelper != null)
				{
					if (proccessWaitHelper.Wait(20))
						return true;
				}
				else
				{
					if (process.WaitForExit(20))
						return true;
				}

				// Wrapper abbrechen, wenn er unterbrochen wurde
				if (handleInterrupt)
				{
					if (interruptWrapperSemWaitHelper != null)
						UserInterrupted = interruptWrapperSemWaitHelper.Wait(1);
					else
						UserInterrupted = InterruptWrapperSem.WaitOne(1);

					if (UserInterrupted)
						break;
				}

				// Zeit ist abgelaufen
				if (haveTimeout && dtWaitUntil < DateTime.Now)
					break;
			}

			return false;
		}

		/// <summary>
		/// Wait for the thread that haved to be end...with nessecary exception handling 
		/// </summary>
		/// <param name="thread">thread which will be wait for</param>
		/// <param name="timeout">time to wait for thread to end</param>
		/// <param name="doAbort">if true, thread will be forced to end, bei sending an abort signal</param>
		/// <param name="infoAbort">Information that will be logged when tread is aborted</param>
		/// <returns></returns>
		private bool JoinAndAbortThread(dynamic thread, int timeout, bool doAbort = false, string infoAbort = "")
		{
			bool waitJoined = false;
			try
			{
				waitJoined = thread == null || thread.Join(timeout);
			}
			catch { }

			if (!waitJoined && doAbort)
			{
				Thread nativeThread = BlockingThread.GetNativeThread(thread);
				ThreadAbortSave(nativeThread, infoAbort);
			}

			return waitJoined;
		}

		/// <summary>
		/// Save aborting a thread
		/// </summary>
		/// <param name="thread"></param>
		/// <param name="info"></param>
		private void ThreadAbortSave(Thread thread, string info)
		{
			try
			{
				if (thread != null && thread.IsAlive)
					thread.Abort(info);
			}
			catch { }
		}

		/// <summary>
		/// Terminates the give GnuPG processes
		/// The agent process is killed und restarted too, if it's needed
		/// </summary>
		/// <param name="gpgWorker"></param>
		/// <param name="gpgPinentry"></param>
		/// <param name="restartAgent"></param>/// 
		/// <returns>true, if the agent process has been restarted</returns>
		/// <!-- Bug 3620 -->
		private bool ResetGPGProcesses(Process gpgWorker, Process gpgPinentry, bool restartAgent)
		{
			try
			{
				bool workerAlreadyStopped = true;
				if (gpgWorker != null && !gpgWorker.HasExited)
					try
					{
						workerAlreadyStopped = false;
						gpgWorker.Kill();
					}
					catch { }

				// No agent process handling if the user has interrupted the GnuPG command
				// Or the GnuPg command has terminated by itself
				if (UserInterrupted || workerAlreadyStopped)
					return false;

				if (!restartAgent)
					return false;

				bool agentKilled = false;
				if (StopAgentDelegate != null)
					agentKilled = StopAgentDelegate(HomeDir, gpgPinentry);

				if (StartAgentDelegate != null && agentKilled)
					return StartAgentDelegate(HomeDir) != null;


			}
			catch { }

			return false;
		}

		/// <summary>
		/// Prüft ob der Pinentry-Prozess läuft und sichtbar ist
		/// </summary>
		/// <param name="isPinentryVisible">Liefert die Sichtbarkeit des Prozesses zurück</param>
		/// <returns>Gibt den gefundenen Pinentryprozess zurück. 
		/// Wurde kein Prozess gefunden wird null zurückgegeben</returns>
		/// /// <!-- Bug 3620 -->
		private Process CheckForPinentry(out bool isPinentryVisible)
		{
			isPinentryVisible = false;

			Process[] processes = Process.GetProcesses();
			foreach (Process proc in processes)
			{
				if (proc == null || string.IsNullOrEmpty(proc.ProcessName))
					continue;

				if (!proc.ProcessName.StartsWith("pinentry", StringComparison.InvariantCultureIgnoreCase))
					continue;

				IntPtr mainWndHandle = FindWindow(proc.Id, "pinentry");
				if (mainWndHandle == IntPtr.Zero)
					continue;

				isPinentryVisible = IsWindowVisible(mainWndHandle);
				return proc;
			}
			return null;
		}

		/// <summary>
		/// Find the caption for the window indentified by hWnd and checks if it starts with the given string argument
		/// </summary>
		/// <param name="hWnd"></param>
		/// <param name="captionStartsWith"></param>
		/// <returns></returns>
		private static bool IsPinEntryWindow(IntPtr hWnd, string captionStartsWith)
		{
			if (hWnd != IntPtr.Zero)
			{
				StringBuilder sb = new StringBuilder(256 + 1);
				if (GetWindowText(hWnd, sb, sb.Capacity) > 0)
					return sb.ToString().StartsWith(captionStartsWith, StringComparison.InvariantCultureIgnoreCase);
			}

			return false;
		}

		/// <summary>
		/// Looking for a window that starts with the given caption of the process
		/// </summary>
		/// <param name="processId"></param>
		/// <param name="captionStartsWith"></param>
		/// <returns></returns>
		private static IntPtr FindWindow(int processId, string captionStartsWith)
		{
			var handles = new List<IntPtr>();
			foreach (ProcessThread thread in Process.GetProcessById(processId).Threads)
				EnumThreadWindows(thread.Id, (hWnd, lParam) => { handles.Add(hWnd); return true; }, IntPtr.Zero);

			foreach (IntPtr handle in handles)
				if (IsPinEntryWindow(handle, captionStartsWith))
					return handle;

			return IntPtr.Zero;
		}

		#endregion
	}
}
